@@ -2,10 +2,11 @@
 # https://github.com/craigahobbs/chisel/blob/master/LICENSE
 
 """
-TODO
+Chisel is a light-weight Python WSGI application framework with tools for building well-documented,
+well-tested, schema-validated JSON web APIs.
 """
 
-__version__ = '0.9.94'
+__version__ = '0.9.96'
 
 from .action import \
     Action, \
