@@ -2,7 +2,7 @@
 # https://github.com/craigahobbs/chisel/blob/master/LICENSE
 
 """
-TODO
+Chisel documentation application
 """
 
 from collections import defaultdict
@@ -12,39 +12,45 @@ from .model import Typedef, TypeStruct, TypeEnum, TypeArray, TypeDict, get_refer
 from .request import RedirectRequest, StaticRequest
 
 
-def create_doc_requests(requests=None, root_path='/doc', request_api=True, doc=True, doc_css=True, cache=True):
+def create_doc_requests(requests=None, root_path='/doc', api=True, app=True, css=True, cache=True):
     """
-    TODO
+    Yield a series of requests for use with :meth:`~chisel.Application.add_requests` comprising the Chisel
+    documentation application. By default, the documenation application is hosted at "/doc/".
 
-    :param requests: A map of request name to :class:`~chisel.Request`.
-    :type requests: dict(str, ~chisel.Request)
-    :param str root_path: TODO
-    :param bool request_api: TODO
-    :param bool doc: TODO
-    :param bool doc_css: TODO
-    :param bool cache: TODO
+    :param requests: A list of requests or None to use the application's requests
+    :type requests: list(~chisel.Request)
+    :param str root_path: The documentation application URL root path. The default is "/doc".
+    :param bool api: If True, include the documentation APIs. Two documentation APIs are added,
+        "/doc/doc_index" and "`/doc/doc_request <doc/doc.html#name=chisel_doc_request>`_".
+    :param bool app: If True, include the documentation client application.
+    :param bool css: If True, and "app" is True, include "/doc/doc.css" and "/doc/doc.svg". If you exclude CSS,
+        you can add your own versions of these requests to customize the documentation styling.
+    :param bool cache: If True, cache static request content.
+    :returns: Generator of :class:`~chisel.Request`
     """
 
-    if request_api:
+    if api:
         yield DocIndex(requests=requests, urls=(('GET', root_path + '/doc_index'),))
-    if request_api:
         yield DocRequest(requests=requests, urls=(('GET', root_path + '/doc_request'),))
-    if doc:
+    if app:
         yield RedirectRequest((('GET', root_path),), root_path + '/')
         yield StaticRequest('chisel', 'static/doc.html', urls=(('GET', root_path + '/'), ('GET', root_path + '/index.html')), cache=cache)
-        yield StaticRequest('chisel', 'static/doc.js', urls=(('GET', root_path + '/doc.js'),), cache=cache)
-        if doc_css:
+        if css:
             yield StaticRequest('chisel', 'static/doc.css', urls=(('GET', root_path + '/doc.css'),), cache=cache)
+            yield StaticRequest('chisel', 'static/doc.svg', urls=(('GET', root_path + '/doc.svg'),), cache=cache)
         yield StaticRequest('chisel', 'static/chisel.js', urls=(('GET', root_path + '/chisel.js'),), cache=cache)
+        yield StaticRequest('chisel', 'static/doc.js', urls=(('GET', root_path + '/doc.js'),), cache=cache)
 
 
 class DocIndex(Action):
     """
-    TODO
+    The documentation index API. This API provides all the information the documentation application needs to render the
+    index page.
 
-    :param requests: A map of request name to :class:`~chisel.Request`.
+    :param requests: A list of requests or None to use the application's requests
     :type requests: dict(str, ~chisel.Request)
-    :param list(tuple) urls: TODO
+    :param list(tuple) urls: The list of URL method/path tuples. The first value is the HTTP request method (e.g. 'GET')
+        or None to match any. The second value is the URL path or None to use the default path.
     """
 
     __slots__ = ('requests',)
@@ -52,38 +58,49 @@ class DocIndex(Action):
     SPEC = '''\
 group "Documentation"
 
-# TODO
+# A non-empty string array
 typedef string[len > 0] StringArray
 
-# TODO
+# Get the documentation index
 action chisel_doc_index
 
     output
 
-        # TODO
+        # The documentation index title
+        string title
+
+        # The dictionary of documentation group titles to array of request names
         StringArray{} groups
 '''
 
     def __init__(self, requests=None, urls=(('GET', '/doc_index'),)):
         super().__init__(self._doc_index, name='chisel_doc_index', urls=urls, spec=self.SPEC)
-        self.requests = requests
+        if requests is not None:
+            self.requests = {request.name: request for request in requests}
+        else:
+            self.requests = None
 
     def _doc_index(self, ctx, unused_req):
+        requests = self.requests if self.requests is not None else ctx.app.requests
         groups = defaultdict(list)
-        for request in (self.requests or ctx.app.requests).values():
+        for request in requests.values():
             groups[request.doc_group or 'Uncategorized'].append(request.name)
         return {
+            'title': ctx.environ['HTTP_HOST'],
             'groups': {group: sorted(names) for group, names in groups.items()}
         }
 
 
 class DocRequest(Action):
     """
-    TODO
+    The documentation request API. This API provides all the information the documentation applicaton needs to render
+    the request documentation page. The documentation request API's documentation is `here
+    <doc/doc.html#name=chisel_doc_request>`_.
 
-    :param requests: A map of request name to :class:`~chisel.Request`.
-    :type requests: dict(str, ~chisel.Request)
-    :param list(tuple) urls: TODO
+    :param requests: A list of requests or None to use the application's requests
+    :type requests: list(~chisel.Request)
+    :param list(tuple) urls: The list of URL method/path tuples. The first value is the HTTP request method (e.g. 'GET')
+        or None to match any. The second value is the URL path or None to use the default path.
     """
 
     __slots__ = ('requests',)
@@ -91,268 +108,260 @@ class DocRequest(Action):
     SPEC = '''
 group "Documentation"
 
-# TODO
-typedef string[len > 0] StringArray
-
-# TODO
-typedef string(len > 0) StructName
-
-# TODO
-typedef string(len > 0) EnumName
-
-# TODO
-typedef string(len > 0) TypedefName
-
-# TODO
+# Union representing a member type
 union Type
 
-    # TODO
+    # A built-in type
     BuiltinType builtin
 
-    # TODO
+    # An array type
     Array array
 
-    # TODO
+    # A dictionary type
     Dict dict
 
-    # TODO
-    EnumName enum
+    # An enumeration type
+    string enum
 
-    # TODO
-    StructName struct
+    # A struct type
+    string struct
 
-    # TODO
-    TypedefName typedef
+    # A type definition
+    string typedef
 
-# TODO
-enum BuiltinType
+# A type attribute
+struct Attribute
 
-    # TODO
-    string
-
-    # TODO
-    int
-
-    # TODO
-    float
-
-    # TODO
-    bool
-
-    # TODO
-    date
-
-    # TODO
-    datetime
-
-    # TODO
-    uuid
-
-    # TODO
-    object
-
-# TODO
-struct Array
-
-    # TODO
-    Type type
-
-    # TODO
-    optional Attr attr
-
-# TODO
-struct Dict
-
-    # TODO
-    Type type
-
-    # TODO
-    optional Attr attr
-
-    # TODO
-    Type key_type
-
-    # TODO
-    optional Attr key_attr
-
-# TODO
-struct Enum
-
-    # TODO
-    optional StringArray doc
-
-    # TODO
-    EnumName name
-
-    # TODO
-    EnumValue[] values
-
-# TODO
-struct EnumValue
-
-    # TODO
-    optional StringArray doc
-
-    # TODO
-    string value
-
-# TODO
-struct Struct
-
-    # TODO
-    optional StringArray doc
-
-    # TODO
-    StructName name
-
-    # TODO
-    optional bool union
-
-    # TODO
-    Member[] members
-
-# TODO
-struct Member
-
-    # TODO
-    optional StringArray doc
-
-    # TODO
-    string name
-
-    # TODO
-    optional bool optional
-
-    # TODO
-    optional bool nullable
-
-    # TODO
-    optional Attr attr
-
-    # TODO
-    Type type
-
-# TODO
-struct Attr
-
-    # TODO
+    # The value is equal
     optional float eq
 
-    # TODO
+    # The value is less than
     optional float lt
 
-    # TODO
+    # The value is less than or equal to
     optional float lte
 
-    # TODO
+    # The value is greater than
     optional float gt
 
-    # TODO
+    # The value is greater than or equal to
     optional float gte
 
-    # TODO
+    # The length is equal to
     optional int len_eq
 
-    # TODO
+    # The length is less-than
     optional int len_lt
 
-    # TODO
+    # The length is less than or equal to
     optional int len_lte
 
-    # TODO
+    # The length is greater than
     optional int len_gt
 
-    # TODO
+    # The length is greater than or equal to
     optional int len_gte
 
-# TODO
-struct Typedef
+# The built-in type enumeration
+enum BuiltinType
 
-    # TODO
-    optional StringArray doc
+    # The string type
+    string
 
-    # TODO
-    TypedefName name
+    # The integer type
+    int
 
-    # TODO
-    optional Attr attr
+    # The float type
+    float
 
-    # TODO
+    # The boolean type
+    bool
+
+    # A date formatted as an ISO-8601 date string
+    date
+
+    # A date/time formatted as an ISO-8601 date/time string
+    datetime
+
+    # A UUID formatted as string
+    uuid
+
+    # An object of any type
+    object
+
+# An array type
+struct Array
+
+    # The contained type
     Type type
 
-# TODO
-struct Action
+    # The contained type's attributes
+    optional Attribute attr
 
-    # TODO
+# A dictionary type
+struct Dict
+
+    # The contained key type
+    Type type
+
+    # The contained key type's attributes
+    optional Attribute attr
+
+    # The contained value type
+    Type key_type
+
+    # The contained value type's attributes
+    optional Attribute key_attr
+
+# An enumeration type
+struct Enum
+
+    # The documentation markdown text
+    optional string doc
+
+    # The enum type name
     string name
 
-    # TODO
+    # The enumeration values
+    EnumValue[] values
+
+# An enumeration type value
+struct EnumValue
+
+    # The documentation markdown text
+    optional string doc
+
+    # The value string
+    string value
+
+# A struct type
+struct Struct
+
+    # The documentation markdown text
+    optional string doc
+
+    # The struct type name
+    string name
+
+    # The struct members
+    Member[] members
+
+    # If true, the struct is a union and exactly one of the optional members is present.
+    optional bool union
+
+# A struct type member
+struct Member
+
+    # The documentation markdown text
+    optional string doc
+
+    # The member name
+    string name
+
+    # The member type
+    Type type
+
+    # The member type attributes
+    optional Attribute attr
+
+    # If true, the member is optional and may not be present.
+    optional bool optional
+
+    # If true, the member may be null.
+    optional bool nullable
+
+# A typedef type
+struct Typedef
+
+    # The documentation markdown text
+    optional string doc
+
+    # The typedef type name
+    string name
+
+    # The typedef's type
+    Type type
+
+    # The typedef's type attributes
+    optional Attribute attr
+
+# A JSON web service API
+struct Action
+
+    # The action name
+    string name
+
+    # The path parameters struct
     Struct path
 
-    # TODO
+    # The query parameters struct
     Struct query
 
-    # TODO
+    # The content body struct
     Struct input
 
-    # TODO
+    # The response body struct
     optional Struct output
 
-    # TODO
+    # The custom error response codes
     Enum errors
 
-# TODO
-struct RequestUrl
+# Struct representing a request's URL information
+struct RequestURL
 
-    # TODO
+    # The request URL HTTP request method. If not present, all HTTP request methods are accepted.
     optional string method
 
-    # TODO
+    # The request URL path
     string url
 
-# TODO
+# Get a request's documentation information
 action chisel_doc_request
 
     query
 
-        # TODO
+        # The request name
         string name
 
     output
 
-        # TODO
-        optional StringArray doc
+        # The documentation markdown text
+        optional string doc
 
-        # TODO
+        # The request name
         string name
 
-        # TODO
-        RequestUrl[] urls
+        # The array of URL paths where the request is hosted.
+        RequestURL[] urls
 
-        # TODO
+        # The action definition. If this member is present then this request is an action.
         optional Action action
 
-        # TODO
+        # The array of struct types referenced by the action
         optional Struct[] structs
 
-        # TODO
+        # The array of enum types referenced by the action
         optional Enum[] enums
 
-        # TODO
+        # The array of typedef types referenced by the action
         optional Typedef[] typedefs
 
     errors
 
-        # TODO
+        # The request name is unknown
         UnknownName
 '''
 
     def __init__(self, requests=None, urls=(('GET', '/doc_request'),)):
         super().__init__(self._doc_request, name='chisel_doc_request', urls=urls, spec=self.SPEC)
-        self.requests = requests
+        if requests is not None:
+            self.requests = {request.name: request for request in requests}
+        else:
+            self.requests = None
 
     def _doc_request(self, ctx, req):
-        request = (self.requests or ctx.app.requests).get(req['name'])
+        requests = self.requests if self.requests is not None else ctx.app.requests
+        request = requests.get(req['name'])
         if request is None:
             raise ActionError('UnknownName')
 
@@ -360,8 +369,8 @@ action chisel_doc_request
             'name': request.name,
             'urls': [self._url_dict(method, url) for method, url in request.urls],
         }
-        if request.doc:
-            response['doc'] = [request.doc] if isinstance(request.doc, str) else request.doc
+        if request.doc is not None:
+            response['doc'] = request.doc
 
         if isinstance(request, Action):
             response['action'] = action_dict = {
@@ -470,8 +479,8 @@ action chisel_doc_request
             'name': struct_type.type_name,
             'members': [cls._member_dict(member) for member in struct_type.members()],
         }
-        if struct_type.doc:
-            struct_dict['doc'] = [struct_type.doc] if isinstance(struct_type.doc, str) else struct_type.doc
+        if struct_type.doc is not None:
+            struct_dict['doc'] = struct_type.doc
         if struct_type.union:
             struct_dict['union'] = True
         return struct_dict
@@ -482,8 +491,8 @@ action chisel_doc_request
             'name': member.name,
             'type': cls._type_dict(member.type),
         }
-        if member.doc:
-            member_dict['doc'] = member.doc if isinstance(member.doc, str) else member.doc
+        if member.doc is not None:
+            member_dict['doc'] = member.doc
         if member.optional:
             member_dict['optional'] = member.optional
         if member.nullable:
@@ -498,8 +507,8 @@ action chisel_doc_request
             'name': enum_type.type_name,
             'values': [cls._enum_value_dict(enum_value) for enum_value in enum_type.values()],
         }
-        if enum_type.doc:
-            enum_dict['doc'] = enum_type.doc if isinstance(enum_type.doc, str) else enum_type.doc
+        if enum_type.doc is not None:
+            enum_dict['doc'] = enum_type.doc
         return enum_dict
 
     @classmethod
@@ -507,8 +516,8 @@ action chisel_doc_request
         enum_value_dict = {
             'value': enum_value.value,
         }
-        if enum_value.doc:
-            enum_value_dict['doc'] = enum_value.doc if isinstance(enum_value.doc, str) else enum_value.doc
+        if enum_value.doc is not None:
+            enum_value_dict['doc'] = enum_value.doc
         return enum_value_dict
 
     @classmethod
@@ -517,8 +526,8 @@ action chisel_doc_request
             'name': typedef_type.type_name,
             'type': cls._type_dict(typedef_type.type),
         }
-        if typedef_type.doc:
-            typedef_dict['doc'] = typedef_type.doc if isinstance(typedef_type.doc, str) else typedef_type.doc
+        if typedef_type.doc is not None:
+            typedef_dict['doc'] = typedef_type.doc
         if typedef_type.attr is not None:
             typedef_dict['attr'] = cls._attr_dict(typedef_type.attr)
         return typedef_dict
