@@ -2,7 +2,7 @@
 # https://github.com/craigahobbs/chisel/blob/master/LICENSE
 
 """
-TODO
+Chisel request base class and common request classes
 """
 
 from functools import partial
@@ -23,9 +23,40 @@ REQUESTS_MODULE_ATTR = '__chisel_requests__'
 
 def request(wsgi_callback=None, **kwargs):
     """
-    TODO
+    Decorator for creating a :class:`~chisel.Request` object that wraps a WSGI application function. For example:
 
-    :param ~collections.abc.Callable wsgi_callback: TODO
+    >>> import chisel
+    ...
+    >>> @chisel.request
+    ... def my_request(environ, start_response):
+    ...    start_response('200 OK', [('Content-Type', 'text/plain')])
+    ...    return [b'Hello, World!']
+    ...
+    >>> my_request.name, my_request.urls, my_request.doc, my_request.doc_group
+    ...
+    ('my_request', ((None, '/my_request'),), None, None)
+
+    You can also pass :class:`~chisel.Request` initialization parameters to the request decorator:
+
+    >>> @chisel.request(urls=[('GET', None)], doc='This is my request')
+    ... def my_request(environ, start_response):
+    ...    start_response('200 OK', [('Content-Type', 'text/plain')])
+    ...    return [b'Hello, World!']
+    ...
+    >>> my_request.name, my_request.urls, my_request.doc, my_request.doc_group
+    ...
+    ('my_request', (('GET', '/my_request'),), 'This is my request', None)
+
+    The created :class:`~chisel.Request` object is passed to an application's :meth:`~chisel.add_request` method to host
+    it with that applicaton.
+
+    >>> application = chisel.Application()
+    >>> application.add_request(my_request)
+    >>> application.request('GET', '/my_request')
+    ('200 OK', [('Content-Type', 'text/plain')], b'Hello, World!')
+
+    :param ~collections.abc.Callable wsgi_callback: A WSGI application function
+    :returns Request: The created Request object
     """
 
     if wsgi_callback is None:
@@ -34,15 +65,15 @@ def request(wsgi_callback=None, **kwargs):
 
 
 class Request:
-    """
-    TODO
+    """The Chisel Request object is a wrapper that associates hosting metadata with a WSGI application function.
+    See the :func:`~chisel.request` decorator for a common way to create Request objects.
 
-    :param ~collections.abc.Callable wsgi_callback: TODO
-    :param str name: TODO
-    :param list(tuple) urls: TODO
-    :param doc: TODO
-    :type doc: list(str) or str
-    :param str doc_group: TODO
+    :param ~collections.abc.Callable wsgi_callback: A WSGI application function
+    :param str name: The request name. The default name is the callback function's name.
+    :param list(tuple) urls: The list of URL method/path tuples. The first value is the HTTP request method (e.g. 'GET')
+        or None to match any. The second value is the URL path or None to use the default path.
+    :param str doc: The documentation markdown text
+    :param str doc_group: The documentation group
     """
 
     __slots__ = ('wsgi_callback', 'name', 'urls', 'doc', 'doc_group')
@@ -50,21 +81,21 @@ class Request:
     def __init__(self, wsgi_callback=None, name=None, urls=None, doc=None, doc_group=None):
         assert wsgi_callback is not None or name is not None, 'must specify either wsgi_callback and/or name'
 
-        #: TODO
+        #: The WSGI application function
         self.wsgi_callback = wsgi_callback
 
-        #: TODO
+        #: The request name
         self.name = name or wsgi_callback.__name__
 
-        #: TODO
+        #: The documentation markdown text
         self.doc = doc
 
-        #: TODO
+        #: The documentation group
         self.doc_group = doc_group
 
         # Normalize urls into list of uppercase-method/path tuple pairs
         if urls is None:
-            #: TODO
+            #: The list of URL method/path tuples
             self.urls = ((None, '/' + self.name),)
         else:
             self.urls = tuple(chain.from_iterable(
@@ -75,7 +106,8 @@ class Request:
 
     def __call__(self, environ, start_response):
         """
-        TODO
+        The request WSGI application method. By default the wrapped wsgi_callback is called. Sub-classes may override this
+        method without calling this method.
 
         :param dict environ: The :pep:`WSGI <3333>` environ dictionary.
         :param ~collections.abc.Callable start_response: The :pep:`WSGI <3333>` start-response callable.
@@ -86,9 +118,10 @@ class Request:
 
     def decorate_module(self, wsgi_callback):
         """
-        TODO
+        Helper method to update a collection requests on the request's module. This information is used by
+        :meth:`~chisel.Request.import_requests`.
 
-        :param ~collections.abc.Callable wsgi_callback: TODO
+        :param ~collections.abc.Callable wsgi_callback: The wrapped WSGI application function
         """
 
         if wsgi_callback.__module__: # pragma: no branch
@@ -103,10 +136,10 @@ class Request:
     @staticmethod
     def import_requests(package, parent_package=None):
         """
-        TODO
+        Import reqeusts from a package.
 
-        :param str package: TODO
-        :param str parent_package: TODO
+        :param str package: The package in which to recursively load requests
+        :param str parent_package: The parent package
         """
 
         package = importlib.import_module(package, parent_package)
@@ -119,15 +152,15 @@ class Request:
 
 class RedirectRequest(Request):
     """
-    TODO
+    A redirect reqeust
 
-    :param list(tuple) urls: TODO
-    :param str redirect_url: TODO
-    :param bool permanent: TODO
-    :param str name: TODO
-    :param doc: TODO
-    :type doc: list(str) or str
-    :param str doc_group: TODO
+    :param list(tuple) urls: The list of URL method/path tuples. The first value is the HTTP request method (e.g. 'GET')
+        or None to match any. The second value is the URL path or None to use the default path.
+    :param str redirect_url: The redirectd URL
+    :param bool permanent: If True, this is a permanent redirect
+    :param str name: The request name. By default the name is "redirect_<redirect_url>".
+    :param str doc: The documentation markdown text
+    :param str doc_group: The documentation group
     """
 
     __slots__ = ('_status', '_redirect_url')
@@ -149,18 +182,18 @@ class RedirectRequest(Request):
 
 class StaticRequest(Request):
     """
-    TODO
+    A static resoruce request
 
-    :param str package: TODO
-    :param str resource_name: TODO
-    :param bool cache: TODO
-    :param str content_type: TODO
-    :param list(tuple) headers: TODO
-    :param str name: TODO
-    :param list(tuple) urls: TODO
-    :param doc: TODO
-    :type doc: list(str) or str
-    :param str doc_group: TODO
+    :param str package: The package containing the static resource
+    :param str resource_name: The resource name (path)
+    :param bool cache: If True, cache the static content. Otherwise, the content is loaded each request.
+    :param str content_type: Optional content type string. If None, the content type is auto-determined.
+    :param list(tuple) headers: Optional list of key/value header tuples
+    :param str name: The request name. By default the name is "static_<redirect_url>".
+    :param list(tuple) urls: The list of URL method/path tuples. The first value is the HTTP request method (e.g. 'GET')
+        or None to match any. The second value is the URL path or None to use the default path.
+    :param str doc: The documentation markdown text
+    :param str doc_group: The documentation group
     """
 
     __slots__ = ('_package', '_resource_name', '_cache', '_headers', '_content', '_etag')
@@ -171,6 +204,7 @@ class StaticRequest(Request):
         '.jpg': 'image/jpeg',
         '.js': 'application/javascript',
         '.png': 'image/png',
+        '.svg': 'image/svg+xml',
         '.txt': 'text/plain',
     }
 
